from pyes.utils import *
# PYJSX
class PYJSX():

    @classmethod
    def jspy_args(self, ss):
        
        #boolean
        if(ss in ['true', 'false']): return ss.title()
        #string
        if(type(ss) == str):         return '\"{0}\"'.format(ss)
        
        #any
        return str(ss)
    
    @classmethod
    def extract_request(self, intf):

        # Super basic request checking:
        pp = intf['active_req']['road']
        ff = intf['active_req']['trac']
        ss = intf['active_req']['seed']
        if(not os.path.exists(pp)): raise ValueError("Python:PYJSX:process_intf: Invalid Road")
        if(not type(ff) == str):    raise ValueError("Python:PYJSX:process_intf: Invalid Trac")
        if(not type(ff) == list):   raise ValueError("Python:PYJSX:process_intf: Invalid Seed")

        return dotdict({

            "path": pp,
            "name": Utils.file_name(pp),
            "func": ff,
            "args": ','.join(self.jspy_args(arg) for arg in ss) 
        })
    
    @classmethod
    def execute_request(self, request):

        # append to sys path
        sys.path.append(os.path.dirname(request.path))

        n = request.name
        n_as = n.replace('.', '_')
        f = request.func
        a = request.args
        # execute and harvest
        try:
            exec('import {0} as {1}'.format(n, n_as))
            result = eval('{name_as}.{func}({args})'.format(
                        name_as = n_as,
                        func = f, args = a )
                    )

        except Exception as e:
            result = 'Python:PYJSX:execute_request:{err}'.format(err = str(e).replace('\'', '\\\''))
        
        return result

    @classmethod
    def execute_intf(self, intf):
        return self.execute_request(
            self.extract_request(intf)
        )


from pyes.utils import *
# File Interface Class
class FileInterface():

    _key_paths = {
        'info/contacts': list,
        'info/requests_made': int,
        'info/requests_exec': int,
        'info/requests_arch': list,
        'active_req/road': str,
        'active_req/trac': str,
        'active_req/seed': list,
        'active_req/crop': str
    }

    _default_value = dotdict({
            "info": {"contacts": [], "requests_arch": [], "requests_made":0, "requests_exec": 0},
            "active_req": {"road": "", "trac": "", "seed": [], "crop": ""}
    })

    _schema_lambs = dotdict({

        "listt" : lambda c: c if (type(c) == list) else [],
        "pintt" : lambda n: n if (type(n) == int and n>0) else 0,
        "pathh" : lambda p: p if os.path.exists(p) else ""
    })

    _default_schema = Schema({

            "info":
            {
                "contacts": Use(_schema_lambs.listt),
                "requests_arch": Use(_schema_lambs.listt),
                "requests_made": Use(_schema_lambs.pintt),
                "requests_exec": Use(_schema_lambs.pintt)
            },
            
            "active_req":
            {
                "road": Use(_schema_lambs.pathh),
                "trac": Use(str),
                "seed": Use(_schema_lambs.listt),
                "crop": Use(lambda s:s)
            }
        }, ignore_extra_keys=True, description= "File Interface Schema")

    def __init__(self, intf_path = "C:/Users/{0}/AppData/Roaming/PYJSX/INTFS/intf1.json".format(Utils.get_user())):

        self.value = self._default_value
        self.path = intf_path

        #if pulled is invalid: recover remains/push
        self.pull(auto_repair=True)
    
    # [VALIDATORS]
    def validate(self, oo):
        return self._default_schema.is_valid(oo)

    # Handle missing and existing args:
    def recover(self, oo):

        try:
            recovered = self._default_schema.validate(oo)
        except:
            recovered = dotdict(self._default_value)
            oo = dotdict(oo)

            for k, v in self._key_paths.items():
                val = oo.find(k)
                if(val and type(val) == v):
                    recovered.update(k, val)
        
        return recovered

    # [GRABBERS]
    def grab_raw(self):
        with open(self.path, 'r') as f: c = f.read()
        return c

    def grab_proper(self):

        if(not os.path.exists(self.path)): return None
        with open(self.path ,'r') as f:
            cc = json.loads(f.read() or "") 
            
        return self.recover(cc)

    def pull(self, auto_repair = False):
        proper_pulled = self.grab_proper() 
        if(proper_pulled is not None): 
            self.value = proper_pulled
            if(auto_repair):
                self.push(new_value= proper_pulled)
            return True
        
        return False

    def grab_signal(self):

        return '{dir_name}/executed_{intf_name}.tmp'.format(
                            dir_name  = os.path.dirname(self.path),
                            intf_name = os.path.basename(self.path)
                    )

    # [SETTERS/MODIFIERS]
    def push(self, new_value = None):
        return Utils.make_file(self.path, cc = json.dumps(
            new_value if self.validate(new_value) else self.value,
            ensure_ascii= False, indent=4  
        ))
    
    def update(self, k_path, val, source = True):

        self.value = dotdict(self.value).update(k_path, val)
        if(source): self.push()


# Utils

# imports:
#   json (data ser/deser)
#   sys  (path append)
#   os   (path operations)
#   schema (Schema detection/validation)
import json, os, sys
from schema import Schema, Use, Optional

class Utils():

    @classmethod
    # make dir and file if they don't exist
    def make_file(self, pp, cc = None): #cc: contents
        ppp = os.path.dirname(pp)
        if(not os.path.exists(ppp)):
            os.makedirs(ppp)
        if(not os.path.exists(pp)):
            with open(pp, 'w') as f: f.write(cc or "")
            return True
        if(cc is not None):
            with open(pp, 'w') as f: f.write(str(cc))
            return True
        return False

    @classmethod
    def file_name(self, pp):
        return '.'.join(pp.split('/')[-1].split('.')[0:-1])

    @classmethod
    def get_user(self):
        return os.path.split(os.path.expanduser('~'))[-1]


class dotdict(dict):
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def update(self, key_path, value):
        keys = key_path.split('/')
        data = 'self'

        for i in range(0, len(keys)):
            data = data + "[\"{0}\"]".format(keys[i])
        
        # preprocess the crop value before
        # hitting "exec"
        if(type(value) == str):
            value = value.replace('\n', '\\n')
            value = "\"{0}\"".format(value)

        data = "{0} = {1}".format(data, value)
        exec(data)

        return self
        
    def find(self, key_path):
        keys = key_path.split('/')
        curr = self

        for i in range(0, len(keys)):
            try: curr = curr[keys[i]]
            except: return None
        return curr